import guidance

def test_role():
    """ Test the behavior of `role`.
    """

    llm = guidance.llms.Mock()
    prompt = guidance("""
{{#role 'system'~}}
You are an assistant.
{{~/role}}

{{#role 'user'~}}
What is the weather?
{{~/role}}

{{#role 'assistant'~}}
{{gen}}
{{~/role}}
""", llm=llm)

    out = prompt()
    assert str(out) == '\n<|im_start|>system\nYou are an assistant.<|im_end|>\n\n<|im_start|>user\nWhat is the weather?<|im_end|>\n\n<|im_start|>assistant\nmock output 0<|im_end|>\n'

def test_short_roles():
    """ Test the behavior of the shorthand versions of `role`.
    """

    llm = guidance.llms.Mock()
    prompt = guidance("""
{{#system~}}
You are an assistant.
{{~/system}}

{{#user~}}
What is the weather?
{{~/user}}

{{#assistant~}}
{{gen}}
{{~/assistant}}
""", llm=llm)

    out = prompt(test="asdfa")
    assert str(out) == '\n<|im_start|>system\nYou are an assistant.<|im_end|>\n\n<|im_start|>user\nWhat is the weather?<|im_end|>\n\n<|im_start|>assistant\nmock output 0<|im_end|>\n'
    


import guidance
import pytest
from ..utils import get_llm

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_select(llm):
    """ Test the behavior of `select`.
    """

    llm = get_llm(llm)
    program = guidance("Is Everest very tall?\nAnswer 'Yes' or 'No': '{{#select 'name'}}Yes{{or}}No{{/select}}", llm=llm)
    out = program()
    assert out["name"] in ["Yes", "No"]

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_select_longtext(llm):
    """ Test the behavior of `select`.
    """

    llm = get_llm(llm)
    program = guidance("""Is Everest very tall?\nAnswer:
{{#select 'name'}}No because of all the other ones.{{or}}Yes because I saw it.{{/select}}""", llm=llm)
    out = program()
    assert out["name"] in ["No because of all the other ones.", "Yes because I saw it."]

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_select_with_list(llm):
    """ Test the behavior of `select` in non-block mode.
    """

    llm = get_llm(llm)
    program = guidance("Is Everest very tall?\nAnswer 'Yes' or 'No': '{{select 'name' options=options logprobs='logprobs'}}", llm=llm)
    out = program(options=["Yes", "No", "Maybe", "I don't know"])
    assert out["name"] in ["Yes", "No", "Maybe", "I don't know"]
    for k in out["logprobs"]:
        assert out["logprobs"][k] <= 0

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_select_list_append(llm):
    """ Test the behavior of `select` with list_append=True.
    """

    llm = get_llm(llm)
    program = guidance("Is Everest very tall?\n{{select 'name' options=options list_append=True}}\n{{select 'name' options=options list_append=True}}", llm=llm)
    out = program(options=["Yes", "No"])
    assert len(out["name"]) == 2
    for v in out["name"]:
        assert v in ["Yes", "No"]

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_select_names(llm):
    """ Test the behavior of `select` with hard to guess terms.
    """

    llm = get_llm(llm)
    out = guidance(
        "Hello, {{#select 'name'}}Alice{{or}}Bob{{/select}}",
        llm=llm
    )()
    assert out["name"] in ["Alice", "Bob"]

def test_select_multi_path():
    """ Test the behavior of `select` and confirm the returns probability distribution sums to 1.
    """
    import numpy as np

    options = [
        "This is one sentence about fish and dogs.",
        "This is another sentence about fish and dogs.",
        "Sure, here is a sentence about cats.",
        "Sure thing, here is a sentence about cats.",
        "This is one"
    ]

    llm = get_llm("transformers:gpt2")
    out = guidance(
        "Hello, write me a sentence. {{select 'sentence' logprobs='probs' options=options}}",
        llm=llm
    )(options=options)
    assert abs(1 - np.exp([l for l in out["probs"].values()]).sum()) < 1e-5
    assert out["sentence"] in options

def test_select_multi_path_with_suffix():
    """ Test the behavior of `select` and confirm the returns probability distribution sums to 1.
    """
    import numpy as np

    options = [
        "This is one sentence about fish and dogs.",
        "This is another sentence about fish and dogs.",
        "Sure, here is a sentence about cats.",
        "Sure thing, here is a sentence about cats.",
        "This is one"
    ]

    llm = get_llm("transformers:gpt2")
    out = guidance(
        "Hello, write me a sentence. {{select 'sentence' logprobs='probs' options=options}} And this is the suffix.",
        llm=llm
    )(options=options)
    assert abs(1 - np.exp([l for l in out["probs"].values()]).sum()) < 1e-5
    assert out["sentence"] in options

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_select_odd_spacing(llm):
    """ Test the behavior of `select` with list_append=True.
    """

    llm = get_llm(llm)
    prompt = guidance('''Is the following sentence offensive? Please answer with a single word, either "Yes", "No", or "Maybe".
    Sentence: {{example}}
    Answer: {{#select "answer" logprobs='logprobs'}} Yes{{or}} Nein{{or}} Maybe{{/select}}''', llm=llm)
    prompt = prompt(example='I hate tacos.')
    assert prompt["answer"] in [" Yes", " Nein", " Maybe"]

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_overlapping_options(llm):
    """ Test the behavior of `select` when one option is a prefix of another.
    """

    llm = get_llm(llm)
    options = ['a', 'aa']
    program = guidance("'{{select options=options}}", llm=llm)
    out = program(options=options)
    assert out["selected"] in options

@pytest.mark.parametrize("llm", ["transformers:gpt2", "openai:text-curie-001"])
def test_non_greedy_tokenize(llm):
    """ Test the behavior of `select` when the GPT tokenizer is not greedy (odd space handling).
    """

    llm = get_llm(llm)
    program = guidance('''Is the following sentence offensive? Please answer with a single word, either "Yes", "No", or "Maybe".
Sentence: {{example}}
Answer:{{#select "answer" logprobs='logprobs'}} 
    Yes{{or}} 
    No{{or}} 
    Maybe
{{/select}}''', llm=llm)
    executed_program = program(example='I hate tacos')
    assert executed_program["answer"] in [" \n    Yes", " \n    No", " \n    Maybe\n"]

import guidance

guidance.llm = guidance.llms.OpenAI("text-davinci-003")
# the {{#select}} command allows you to use the LLM to select from a set of options


# TODO: fix this next
# def test_unexpected_tokens():
#     """ Test the behavior of `select` when the next tokens are hard to predict.
#     """

#     llm = get_transformers_llm("gpt2")
#     options = ['a', 'b']
#     program = guidance("some word xy{{select options=options}}", llm=llm)
#     out = program(options=options)
#     assert out["selected"] in options

# TODO: test when we have few starting tokens


import parsimonious

# define the Guidance language grammar
grammar = parsimonious.grammar.Grammar(
r"""
template = template_chunk*
template_chunk = comment / slim_comment / escaped_command / unrelated_escape / command / command_block / content

comment = comment_start comment_content* comment_end
comment_start = "{{!--"
comment_content = not_comment_end / ~r"[^-]*"
not_comment_end = "-" !"-}}"
comment_end = "--}}"

slim_comment = slim_comment_start slim_comment_content* slim_comment_end
slim_comment_start = "{{" "~"? "!"
slim_comment_content = not_slim_comment_end / ~r"[^}]*"
not_slim_comment_end = "}" !"}"
slim_comment_end = "}}"

command = command_start command_content command_end
command_block = command_block_open template (command_block_sep template)* command_block_close
command_block_open = command_start "#" block_command_call command_end
command_block_sep = command_start ("or" / "else") command_end
command_block_close = command_start "/" command_name command_end
command_start = "{{" !"!" "~"?
not_command_start = "{" !"{"
not_command_escape = "\\" !"{{"
command_end = "~"? "}}"
command_contents = ~'[^{]*'
block_command_call = command_name command_args?
command_content = command_call / variable_ref
command_call = command_name command_args
command_args = command_arg_and_ws+
command_arg_and_ws = ws command_arg
command_arg = named_command_arg / positional_command_arg
positional_command_arg = command_arg_group / literal / variable_ref
named_command_arg = variable_name "=" (literal / variable_ref)
command_arg_group = "(" command_content ")"
ws = ~r'\s+'
command_contentasdf = ~"[a-z 0-9]*"i
command_name = ~r"[a-z][a-z_0-9\.]*"i / "<" / ">" / "==" / "!=" / ">=" / "<="
variable_ref = !"or" !"else" ~r"[@a-z][a-z_0-9\.\[\]\"'-]*"i
variable_name = ~r"[@a-z][a-z_0-9]*"i
contentw = ~r'.*'
content = (not_command_start / not_command_escape / ~r"[^{\\]")+
unrelated_escape = "\\" !command_start
escaped_command = "\\" command_start ~r"[^}]*" command_end

literal = string_literal / number_literal / boolean_literal / array_literal / object_literal

string_literal = ~r'"[^\"]*"' / ~r"'[^\']*'"

number_literal = ~r"[0-9\.]+"

boolean_literal = "True" / "False"

array_literal = empty_array / single_item_array / multi_item_array
empty_array = array_start ws? array_end
single_item_array = array_start ws? array_item ws? array_end
array_sep = ws? "," ws?
multi_item_array = array_start ws? array_item (array_sep array_item)* ws? array_end
array_start = "["
array_end = "]"
array_item = literal

object_literal = empty_object / single_item_object / multi_item_object
empty_object = object_start ws? object_end
single_item_object = object_start ws? object_item ws? object_end
object_sep = ws? "," ws?
multi_item_object = object_start ws? object_item (object_sep object_item)* ws? object_end
object_start = "{"
object_end = "}"
object_item = string_literal ws? ":" ws? literal
""")


import os
import pathlib
import diskcache
import asyncio
import platformdirs

class LLM():

    cache_version = 1
    default_system_prompt = "You are a helpful assistant."

    def __init__(self):
        self.chat_mode = False # by default models are not in role-based chat mode
        self.model_name = "unknown"
    
    def __call__(self, *args, asynchronous=False, **kwargs):
        """ Creates a session and calls the LLM with the given arguments.

        Note that this is a convenience wrapper so you don't have to call session(),
        for higher performance across multiple calls, use a session directly.
        """
        with self.session(asynchronous=asynchronous) as s:
            out = s(*args, **kwargs)
        return out
    
    def __getitem__(self, key):
        """ Gets an attribute from the LLM.
        """
        return getattr(self, key)
    
    def session(self, asynchronous=False):
        """ Creates a session for the LLM.

        This implementation is meant to be overridden by subclasses.
        """
        
        if asynchronous:
            return LLMSession(self)
        else:
            return SyncSession(LLMSession(self))
    
    def encode(self, string, **kwargs):
        return self._tokenizer.encode(string, **kwargs)
    
    def decode(self, tokens, **kwargs):
        return self._tokenizer.decode(tokens, **kwargs)
    
    @staticmethod
    def _open_cache(file_name):
        return diskcache.Cache(os.path.join(platformdirs.user_cache_dir("guidance"), file_name))
    

class LLMSession():
    def __init__(self, llm):
        self.llm = llm
        self._call_counts = {} # used to track the number of repeated identical calls to the LLM with non-zero temperature
    
    def __enter__(self):
        return self
    
    async def __call__(self, *args, **kwargs):
        return self.llm(*args, **kwargs)

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def _gen_key(self, args_dict):
        var_names = list(args_dict.keys())[1:] # skip the "self" arg
        return "_---_".join([str(v) for v in ([args_dict[k] for k in var_names] + [self.llm.model_name, self.llm.__class__.__name__, self.llm.cache_version])])
    
    def _cache_key(self, args_dict):
        """ Get a cache key for the given args.
        """

        # generate the key without the call count included
        key = self._gen_key(args_dict)

        # if we have non-zero temperature we include the call count in the cache key
        if args_dict.get("temperature", 0) > 0:
            args_dict["call_count"] = self._call_counts.get(key, 0)

            # increment the call count
            self._call_counts[key] = args_dict["call_count"] + 1

            # regenerate the key with the call count in it
            key = self._gen_key(args_dict)

        return key

class SyncSession():
    def __init__(self, session):
        self._session = session
    
    def __enter__(self):
        self._session.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        return self._session.__exit__(exc_type, exc_value, traceback)
    
    def __call__(self, *args, **kwargs):
        return asyncio.get_event_loop().run_until_complete(self._session.__call__(*args, **kwargs))


import guidance

def test_unless():
    """ Test the behavior of `unless`.
    """

    program = guidance("""Answer: {{#unless flag}}Yes{{/unless}}""")

    for flag in [True, 1, "random text"]:
        out = program(flag=flag)
        assert str(out) == "Answer: "
    
    for flag in [False, 0, ""]:
        out = program(flag=flag)
        assert str(out) == "Answer: Yes"


__version__ = "0.0.54"

import types
import sys
import os
import requests
from ._program import Program
from . import llms
from . import library
from ._utils import load, chain
from . import selectors
import nest_asyncio
import asyncio

# the user needs to set an LLM before they can use guidance
llm = None

# This makes the guidance module callable
class Guidance(types.ModuleType):
    def __call__(self, template, llm=None, cache_seed=0, logprobs=None, silent='auto', async_mode=False, stream=None, caching=None, await_missing=False, **kwargs):
        return Program(template, llm=llm, cache_seed=cache_seed, logprobs=logprobs, silent=silent, async_mode=async_mode, stream=stream, caching=caching, await_missing=await_missing, **kwargs)
sys.modules[__name__].__class__ = Guidance


def load(guidance_file):
    ''' Load a guidance program from the given text file.

    If the passed file is a valid local file it will be loaded directly.
    Otherwise, if it starts with "http://" or "https://" it will be loaded
    from the web.
    '''

    if os.path.exists(guidance_file):
        with open(guidance_file, 'r') as f:
            template = f.read()
    elif guidance_file.startswith('http://') or guidance_file.startswith('https://'):
        template = requests.get(guidance_file).text
    else:
        raise ValueError('Invalid guidance file: %s' % guidance_file)
    
    return sys.modules[__name__](template)



import getpass
import sys
import subprocess
import shlex
import os
import subprocess
import asyncio


def shell(command, safe=True, _parser_context=None):
    """ Execute a shell command on the local machine (with user confirmation by default).

    Parameters
    ----------
    command : str
        The command to execute.
    safe : bool
        If True, the user will be asked to confirm the command before it is executed.
    """
    
    partial_output = _parser_context['partial_output']
    partial_output("{{execute '"+command+"'}}")
    
    # before running the command we need to get confirmation from user through keyboard input
    # this is to prevent accidental execution of commands
    if safe:
        c = ""
        while c != "\r" and c != "\n":
            c = getch()
            if c == '\x03':
                raise KeyboardInterrupt()
            elif c == "\x1b":
                partial_output("\nABORTED BY USER\n")
                return "{{execute '"+command+"'}}\n"+"ABORTED BY USER\n"

    # run the command
    all_output = "\n"
    partial_output(all_output)
    try:
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        while True:
            output = process.stdout.readline().decode("utf-8") 
            all_output += output
            partial_output(output)
            if process.poll() is not None:
                break
        rc = process.poll()
        if rc != 0:
            raise Exception("Command failed")
    except Exception as e:
        all_output += str(e)+"\n"
        partial_output(str(e)+"\n")

    return "{{execute '"+command+"'}}"+all_output


# https://stackoverflow.com/questions/510357/how-to-read-a-single-character-from-the-user
class _Getch:
    """Gets a single character from standard input.  Does not echo to the
screen."""
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self): return self.impl()


class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()


getch = _Getch()


class Shell:
    """ A stateful shell object can be used to execute commands and get the output.
    """

    def __init__(self):
        import pty # TODO: Make this work on Windows
        self.shell_cmd = os.environ.get('SHELL', '/bin/sh')
        self.master_fd, self.slave_fd = pty.openpty()
        my_env = os.environ.copy()
        self.p = subprocess.Popen(
            self.shell_cmd,
            preexec_fn=os.setsid,
            stdin=self.slave_fd,
            stdout=self.slave_fd,
            stderr=self.slave_fd,
            text=True,
            env=my_env
        )
        self._env_inited = False
        self.loop = asyncio.get_event_loop()
        self.loop.add_reader(self.master_fd, self.read)

        # an event that the command has finished
        self.command_finished = asyncio.Event()
        self._current_output = ""

    def read(self):
        """ Read the next chunk of output from the shell.
        """
        new_text = os.read(self.master_fd, 10240).decode()
        self._current_output += new_text
        if self._current_output.endswith('autosh$ '):
            self.command_finished.set()
    
    async def __call__(self, command):
        ''' Send a command to the shell and return the output.
        '''

        # set up the environment if it hasn't been done yet
        if not self._env_inited:
            self._env_inited = True
            await self('export set PS1="autosh$ "')
        
        # send the command to the shell
        command = command + '\n'
        os.write(self.master_fd, command.encode())

        # wait for the command to finish
        await self.command_finished.wait()
        self.command_finished.clear()

        # return the collected output
        out = self._current_output
        self._current_output = ""
        return out


import guidance

def test_contains():
    """ Test the behavior of `contains`.
    """

    program = guidance("""{{#if (contains val "substr")}}are equal{{else}}not equal{{/if}}""")
    assert str(program(val="no sub")) == "not equal"
    assert str(program(val="this is a substr")) == "are equal"
    assert str(program(val="this is a subsr")) == "not equal"


import guidance

def test_break_each():
    """ Test the behavior of `break` in an `each` loop.
    """

    llm = guidance.llms.Mock()
    program = guidance("""Loop to ten:
{{~#each list}}
{{this}}
{{~#if (equal this 5)}}{{break}}{{/if~}}
{{/each}}""", llm=llm)
    out = program(list=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    assert out.text == "Loop to ten:\n1\n2\n3\n4\n5"


import guidance
from .utils import get_openai_llm

def test_variable_interpolation():
    """ Test variable interpolation in prompt
    """

    prompt = guidance("Hello, {{name}}!")
    assert str(prompt(name="Guidance")) == "Hello, Guidance!"

def test_chat_stream():
    """ Test the behavior of `stream=True` for an openai chat endpoint.
    """

    import asyncio
    loop = asyncio.new_event_loop()

    guidance.llm = get_openai_llm("gpt-3.5-turbo")

    async def f():
        chat = guidance("""<|im_start|>system
You are a helpful assistent.
<|im_end|>
<|im_start|>user
{{command}}
<|im_end|>
<|im_start|>assistant
{{gen 'answer' max_tokens=10}}""", stream=True)
        out = await chat(command="How do I create a Fasttokenizer with hugging face auto?")
        assert len(out["answer"]) > 0
    loop.run_until_complete(f())

def test_chat_display():
    """ Test the behavior of `stream=True` for an openai chat endpoint.
    """

    import asyncio
    loop = asyncio.new_event_loop()

    guidance.llm = get_openai_llm("gpt-3.5-turbo")

    async def f():
        chat = guidance("""<|im_start|>system
You are a helpful assistent.
<|im_end|>
<|im_start|>user
{{command}}
<|im_end|>
<|im_start|>assistant
{{gen 'answer' max_tokens=10}}""")
        out = await chat(command="How do I create a Fasttokenizer with hugging face auto?")
        assert len(out["answer"]) > 0
    loop.run_until_complete(f())

def test_agents():
    """Test agentes, calling prompt twice"""

    guidance.llm = get_openai_llm("gpt-3.5-turbo")

    prompt = guidance('''<|im_start|>system
You are a helpful assistant.<|im_end|>
{{#geneach 'conversation' stop=False}}
<|im_start|>user
{{set 'this.user_text' (await 'user_text')}}<|im_end|>
<|im_start|>assistant
{{gen 'this.ai_text' n=1 temperature=0 max_tokens=900}}<|im_end|>{{/geneach}}''', echo=True)
    prompt = prompt(user_text='Hi there')
    assert len(prompt['conversation']) == 2
    prompt = prompt(user_text='Please help')
    assert len(prompt['conversation']) == 3


import os
import atexit
import json
import platformdirs
from ._openai import OpenAI


class AzureOpenAI(OpenAI):
    """ Azure OpenAI style integration.

    Warning: This class is not finalized and may change in the future.
    """

    cache = OpenAI._open_cache("_azure_openai.diskcache")

    def __init__(self, model=None, client_id=None, authority=None, caching=True, max_retries=5, max_calls_per_min=60, token=None,
                 endpoint=None, scopes=None, temperature=0.0, chat_mode="auto"):
        

        assert endpoint is not None, "An endpoint must be specified!"
        
        # build a standard OpenAI LLM object
        super().__init__(
            model=model, caching=caching, max_retries=max_retries, max_calls_per_min=max_calls_per_min,
            token=token, endpoint=endpoint, temperature=temperature, chat_mode=chat_mode
        )

        self.client_id = client_id
        self.authority = authority
        self.scopes = scopes

        from msal import PublicClientApplication, SerializableTokenCache
        self._token_cache = SerializableTokenCache()
        self._token_cache_path = os.path.join(platformdirs.user_cache_dir("guidance"), "_azure_openai.token")
        self._app = PublicClientApplication(client_id=self.client_id, authority=self.authority, token_cache=self._token_cache)
        if os.path.exists(self._token_cache_path):
            self._token_cache.deserialize(open(self._token_cache_path, 'r').read())

        self._rest_headers["X-ModelType"] = self.model_name

    @property
    def token(self):
        return self._get_token()
    @token.setter
    def token(self, value):
        pass # ignored for now

    def _get_token(self):
        accounts = self._app.get_accounts()
        result = None

        if accounts:
            # Assuming the end user chose this one
            chosen = accounts[0]

            # Now let's try to find a token in cache for this account
            result = self._app.acquire_token_silent(self.scopes, account=chosen)
    
        if not result:
            # So no suitable token exists in cache. Let's get a new one from AAD.
            flow = self._app.initiate_device_flow(scopes=self.scopes)

            if "user_code" not in flow:
                raise ValueError(
                    "Fail to create device flow. Err: %s" % json.dumps(flow, indent=4))

            print(flow["message"])

            result = self._app.acquire_token_by_device_flow(flow)

            # save the aquired token
            with open(self._token_cache_path, "w") as f:
                f.write(self._token_cache.serialize())

        return result["access_token"]


import guidance

def test_hidden_block():
    """ Test the behavior of `if` with an `else` clause.
    """

    prompt = guidance("""This is a test {{#block hidden=True}}example{{/block}}""")
    out = prompt()
    assert out.text == "This is a test "


async def await_(name, _parser_context=None):
    ''' Awaits a variable by returning its value and then deleting it.

    Note that this is useful for repeatedly getting values since programs
    will pause when they need a value that is not yet set. This means
    that putting `await` in a loop will create a stateful "agent" that can
    repeatedly await values when called multiple times.

    Parameters
    ----------
    name : str
        The name of the variable to await.
    '''

    # stop the program completion if we are waiting for a value to be set
    # this will result in a partially completed program that we can then finish
    # later (by calling it again with the variable we need)
    parser = _parser_context['parser']
    if name not in parser.program:
        parser.executing = False
    else:
        value = parser.program[name]
        del parser.program[name]
        return value
    
    # cache = parser.program._await_cache
    # while name not in cache:
    #     parser.program.finish_execute() # allow the program to finish the current call (since we're waiting for a value from the next call now)
    #     # TODO: instead of waiting here, we should just single we are stopping the program completion here
    #     #       and then let all the containing elements record their state into a new program string that
    #     #       we can then use to continue the program completion later in a new object.
    #     cache.update(await parser.program._await_queue.get())
    #     pass
    # value = cache[name]
    # del cache[name]
    # return value


import guidance

def test_assistant():
    """ Basic test of `assistant`.
    """

    llm = guidance.llms.Mock("the output")

    program = guidance("""
{{#system}}You are fake.{{/system}}
{{#user}}You are real.{{/user}}
{{#assistant}}{{gen 'output' save_prompt='prompt'}}{{/assistant}}""", llm=llm)
    out = program()
    assert out["output"] == "the output"
    #assert str(out) == '\n<|im_start|>system\nYou are fake.<|im_end|>\n<|im_start|>user\nYou are real.<|im_end|>\n<|im_start|>assistant\nthe output<|im_end|>'


import guidance

def test_less():
    """ Test the behavior of `less`.
    """

    program = guidance("""{{#if (less val 5)}}less{{else}}not less{{/if}}""")
    assert str(program(val=6)) == "not less"
    assert str(program(val=4)) == "less"
    assert str(program(val=4.3)) == "less"

def test_less_with_symbol():
    """ Test the behavior of `less` used as `<`.
    """

    program = guidance("""{{#if (< val 5)}}less{{else}}not less{{/if}}""")
    assert str(program(val=6)) == "not less"
    assert str(program(val=4)) == "less"
    assert str(program(val=4.3)) == "less"


import guidance

def test_subtract():
    """ Basic test of `subtract`.
    """

    program = guidance("""Write a number: {{set 'user_response' (subtract 20 variable)}}""")
    assert program(variable=10)["user_response"] == 10
    assert abs(program(variable=20.1)["user_response"] + 0.1) < 1e-5


import guidance

def test_await():
    """ Test the behavior of `await`.
    """

    prompt = guidance("""Is Everest very tall?
User response: '{{set 'user_response' (await 'user_response')}}'""")
    waiting_prompt = prompt()
    assert str(waiting_prompt) == "Is Everest very tall?\nUser response: '{{set 'user_response' (await 'user_response')}}'"
    out = waiting_prompt(user_response="Yes")
    assert str(out) == "Is Everest very tall?\nUser response: 'Yes'"


import os
import requests
import inspect
import time
import json
import re
import datetime
import asyncio

log_file = open("log.txt", "a")
def log(*args):
    # print(*args)
    print(datetime.datetime.now().strftime("%H:%M:%S"), *args, file=log_file)
    log_file.flush()

def load(guidance_file):
    ''' Load a guidance program from the given text file.

    If the passed file is a valid local file it will be loaded directly.
    Otherwise, if it starts with "http://" or "https://" it will be loaded
    from the web.
    '''

    if os.path.exists(guidance_file):
        with open(guidance_file, 'r') as f:
            return f.read()
    elif guidance_file.startswith('http://') or guidance_file.startswith('https://'):
        return requests.get(guidance_file).text
    else:
        raise ValueError('Invalid guidance file: %s' % guidance_file)
    
def chain(programs, **kwargs):
    ''' Chain together multiple programs into a single program.
    
    This merges them into a single program like: {{>program1 hidden=True}}{{>program2 hidden=True}}
    '''

    from ._program import Program

    new_template = "".join(["{{>program%d hidden=True}}" % i for i in range(len(programs))])
    for i, program in enumerate(programs):
        if isinstance(program, Program):
            kwargs["program%d" % i] = program
        else:
            sig = inspect.signature(program)
            args = ""
            for name,_ in sig.parameters.items():
                args += f" {name}={name}"
            fname = find_func_name(program, kwargs)
            kwargs["program%d" % i] = Program("{{set (%s%s)}}" % (fname, args), **{fname: program})
            # kwargs.update({f"func{i}": program})
    return Program(new_template, **kwargs)

def find_func_name(f, used_names):
    if hasattr(f, "__name__"):
        prefix = f.__name__.replace("<", "").replace(">", "")
    else:
        prefix = "function"
    
    if prefix not in used_names:
        return prefix
    else:
        for i in range(100):
            fname = f"{prefix}{i}"
            if fname not in used_names:
                return fname

def strip_markers(s):
    return re.sub(r"{{!--G.*?--}}", r"", s, flags=re.MULTILINE | re.DOTALL)

class JupyterComm():
    def __init__(self, target_id, ipython_handle, callback=None, on_open=None, mode="register"):
        from ipykernel.comm import Comm
        
        self.target_name = "guidance_interface_target_"+target_id
        # print("TARGET NAME", self.target_name)
        self.callback = callback
        self.jcomm = None
        self.ipython_handle = ipython_handle
        self.addd = 1
        self.send_queue = asyncio.Queue()
        self.open_event = asyncio.Event()
        self.is_open = False
        asyncio.get_event_loop().create_task(self._send_loop())
        if mode == "register":
            #log("REGISTERING", self.target_name)
            # asyncio.get_event_loop().create_task(self._register())
            def comm_opened(comm, open_msg):
                #log("OPENED")
                self.addd = 2
                self.jcomm = comm
                self.is_open = True
                self.jcomm.on_msg(self._fire_callback)
                self.open_event.set()
                self._fire_callback({"content": {"data": {"event": "opened"}}})
            self.ipython_handle.kernel.comm_manager.register_target(self.target_name, comm_opened)
            # get_ipython().kernel.comm_manager.register_target(self.target_name, comm_opened) # noqa: F821
        elif mode == "open":
            #log("OPENING", self.target_name)
            self.jcomm = Comm(target_name=self.target_name)
            self.jcomm.on_msg(self._fire_callback)
            # self._fire_callback({"content": {"data": "opened"}})
        else:
            raise Exception("Passed mode must be either 'open' or 'register'!")
        
    # async def _register(self):
    #     def comm_opened(comm, open_msg):
    #         #log("OPENED")
    #         self.addd = 2
    #         self.jcomm = comm
    #         self.jcomm.on_msg(self._fire_callback)
    #         self.open_event.set()
    #         self._fire_callback({"content": {"data": {"event": "opened"}}})
    #     get_ipython().kernel.comm_manager.register_target(self.target_name, comm_opened)

    # def send(self, data, wait=False):
    #     self.send_queue.append(data)
    #     if self.jcomm is None:
    #         return
    #     for d in self.send_queue:
    #         self.jcomm.send({"data": json.dumps(d)})
    #     self.send_queue = []

    def clear_send_queue(self):
        while not self.send_queue.empty():
            self.send_queue.get_nowait()
            self.send_queue.task_done()

    def _fire_callback(self, msg):
        self.callback(msg["content"]["data"])

    def send(self, data):
        self.send_queue.put_nowait(data)

    async def _send_loop(self):
        while True:
            #log("SENDING_LOOP")
            if self.jcomm is None:
                self.open_event.clear()
                await self.open_event.wait()
            data = await self.send_queue.get()
            #log("SENDING_LOOP got one!")
            self.jcomm.send({"data": json.dumps(data)})
    
    # async def _waiting_send(self, data):
    #     #log("SENDING", self.jcomm, data)
        
    #     # await the open event if needed
    #     if self.jcomm is None:
    #         self.open_event.clear()
    #         await self.open_event.wait()
    #     #log("SENDING_now", self.jcomm, data)
    #     self.jcomm.send({"data": json.dumps(data)}) # we encode the JSON so iPython doesn't mess it up

# https://stackoverflow.com/questions/15411967/how-can-i-check-if-code-is-executed-in-the-ipython-notebook
def is_interactive():
    import __main__ as main
    return not hasattr(main, '__file__')

def escape_template_block(text):
    return text.replace("$", "&#36;").replace("{", "&#123;").replace("}", "&#125;")

def unescape_template_block(text):
    return text.replace("&#36;", "$").replace("&#123;", "{").replace("&#125;", "}")


import os
import re
import codecs
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), "r") as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name="guidance",
    version=find_version("guidance", "__init__.py"),
    url="https://github.com/slundberg/guidance",
    author="Scott Lundberg and Marco Tulio Ribeiro",
    author_email="scott.lundberg@microsoft.com",
    description="A guidance language for controlling large language models.",
    long_description="Guidance enables you to control modern language models more effectively and efficiently than traditional prompting or chaining. Guidance programs allow you to interleave generation, prompting, and logical control into a single continuous flow matching how the language model actually processes the text.",
    packages=find_packages(exclude=["user_studies", "notebooks", "client"]),
    package_data={"guidance": ["resources/*"]},
    install_requires=[
        "diskcache",
        "openai>=0.27",
        "parsimonious",
        "pygtrie",
        "platformdirs",
        "tiktoken>=0.3",
        "nest_asyncio",
        "requests",
        "numpy"
    ],
    extras_require={
        'docs': [
            'ipython',
            'numpydoc',
            'sphinx_rtd_theme',
            'sphinx',
            'nbsphinx'
        ],
        'test': [
            'pytest',
            'transformers',
            'torch'
        ]
    }
)



import os
import time
import collections
import regex
import pygtrie
import queue
import threading
import logging
from .._llm import LLM, LLMSession, SyncSession
from .._transformers import Transformers, TransformersSession

class LLaMA(Transformers):
    """ A HuggingFace transformers version of the LLaMA language model with Guidance support.
    """

    cache = LLM._open_cache("_llama.diskcache")

    def __init__(self, model, tokenizer=None, device_map=None, **kwargs):
        """ Create a new LLaMA model.
        """

        # load the LLaMA specific tokenizer and model
        import transformers
        if isinstance(model, str):
            if tokenizer is None:
                tokenizer = transformers.LlamaTokenizer.from_pretrained(model, device_map=device_map)
            model = transformers.LlamaForCausalLM.from_pretrained(model, device_map=device_map)

        super().__init__(model, tokenizer=tokenizer, device_map=device_map, **kwargs)


def break_():
    ''' Breaks out of the current loop.

    This is useful for breaking out of a geneach loop early, typically this is used
    inside an `{{#if ...}}...{{/if}}` block.
    '''
    raise StopIteration()